import configparser
import logging
import os
import re
import tempfile
from typing import (
    Optional,
    TYPE_CHECKING,
)

from markupsafe import escape
from sqlalchemy import (
    delete,
    false,
    select,
)
from sqlalchemy.orm import joinedload

import tool_shed.dependencies.repository
from galaxy import (
    util,
    web,
)
from galaxy.tool_shed.util.repository_util import (
    create_or_update_tool_shed_repository,
    extract_components_from_tuple,
    generate_tool_shed_repository_install_dir,
    get_absolute_path_to_file_in_repository,
    get_installed_repository,
    get_installed_tool_shed_repository,
    get_prior_import_or_install_required_dict,
    get_repo_info_tuple_contents,
    get_repository_admin_role_name,
    get_repository_and_repository_dependencies_from_repo_info_dict,
    get_repository_dependency_types,
    get_repository_for_dependency_relationship,
    get_repository_ids_requiring_prior_import_or_install,
    get_repository_owner,
    get_repository_owner_from_clone_url,
    get_tool_shed_from_clone_url,
    get_tool_shed_repository_by_id,
    get_tool_shed_status_for_installed_repository,
    is_tool_shed_client,
    repository_was_previously_installed,
    set_repository_attributes,
)
from tool_shed.util.common_util import generate_clone_url_for
from tool_shed.util.hg_util import (
    changeset2rev,
    create_hgrc_file,
    get_hgrc_path,
    init_repository,
)
from tool_shed.util.metadata_util import (
    get_next_downloadable_changeset_revision,
    get_repository_metadata_by_changeset_revision,
    repository_metadata_by_changeset_revision,
)
from tool_shed.webapp import model
from tool_shed.webapp.model.db import (
    get_repository_by_name_and_owner,
    get_repository_query,
)

if TYPE_CHECKING:
    from sqlalchemy.orm import scoped_session

    from tool_shed.context import (
        ProvidesRepositoriesContext,
        ProvidesUserContext,
    )
    from tool_shed.structured_app import ToolShedApp


log = logging.getLogger(__name__)

VALID_REPOSITORYNAME_RE = re.compile(r"^[a-z0-9\_]+$")


def get_repository_by_id(app: "ToolShedApp", id):
    """Get a repository from the database via id."""
    sa_session = app.model.session
    return sa_session.query(model.Repository).get(app.security.decode_id(id))


def get_role_by_id(app: "ToolShedApp", role_id):
    """Get a Role from the database by id."""
    sa_session = app.model.session
    return sa_session.query(model.Role).get(app.security.decode_id(role_id))


def create_repo_info_dict(
    app: "ToolShedApp",
    repository_clone_url,
    changeset_revision,
    ctx_rev,
    repository_owner,
    repository_name=None,
    repository=None,
    repository_metadata=None,
    tool_dependencies=None,
    repository_dependencies=None,
    trans=None,
):
    """
    Return a dictionary that includes all of the information needed to install a repository into a local
    Galaxy instance.  The dictionary will also contain the recursive list of repository dependencies defined
    for the repository, as well as the defined tool dependencies.

    This method is called from Galaxy under four scenarios:
    1. During the tool shed repository installation process via the tool shed's get_repository_information()
    method.  In this case both the received repository and repository_metadata will be objects, but
    tool_dependencies and repository_dependencies will be None.
    2. When getting updates for an installed repository where the updates include newly defined repository
    dependency definitions.  This scenario is similar to 1. above. The tool shed's get_repository_information()
    method is the caller, and both the received repository and repository_metadata will be objects, but
    tool_dependencies and repository_dependencies will be None.
    3. When a tool shed repository that was uninstalled from a Galaxy instance is being reinstalled with no
    updates available.  In this case, both repository and repository_metadata will be None, but tool_dependencies
    and repository_dependencies will be objects previously retrieved from the tool shed if the repository includes
    definitions for them.
    4. When a tool shed repository that was uninstalled from a Galaxy instance is being reinstalled with updates
    available.  In this case, this method is reached via the tool shed's get_updated_repository_information()
    method, and both repository and repository_metadata will be objects but tool_dependencies and
    repository_dependencies will be None.
    """
    repo_info_dict = {}
    repository = get_repository_by_name_and_owner(app.model.context, repository_name, repository_owner)
    if app.name == "tool_shed":
        # We're in the tool shed.
        repository_metadata = repository_metadata_by_changeset_revision(app.model, repository.id, changeset_revision)
        if repository_metadata:
            metadata = repository_metadata.metadata
            if metadata:
                if trans is not None:
                    tool_shed_url = trans.repositories_hostname
                else:
                    tool_shed_url = web.url_for("/", qualified=True).rstrip("/")
                rb = tool_shed.dependencies.repository.relation_builder.RelationBuilder(
                    app, repository, repository_metadata, tool_shed_url, trans=trans
                )
                # Get a dictionary of all repositories upon which the contents of the received repository depends.
                repository_dependencies = rb.get_repository_dependencies_for_changeset_revision()
                tool_dependencies = metadata.get("tool_dependencies", {})
    if tool_dependencies:
        new_tool_dependencies = {}
        for dependency_key, requirements_dict in tool_dependencies.items():
            if dependency_key in ["set_environment"]:
                new_set_environment_dict_list = []
                for set_environment_dict in requirements_dict:
                    set_environment_dict["repository_name"] = repository_name
                    set_environment_dict["repository_owner"] = repository_owner
                    set_environment_dict["changeset_revision"] = changeset_revision
                    new_set_environment_dict_list.append(set_environment_dict)
                new_tool_dependencies[dependency_key] = new_set_environment_dict_list
            else:
                requirements_dict["repository_name"] = repository_name
                requirements_dict["repository_owner"] = repository_owner
                requirements_dict["changeset_revision"] = changeset_revision
                new_tool_dependencies[dependency_key] = requirements_dict
        tool_dependencies = new_tool_dependencies
    repo_info_dict[repository.name] = (
        repository.description,
        repository_clone_url,
        changeset_revision,
        ctx_rev,
        repository_owner,
        repository_dependencies,
        tool_dependencies,
    )
    return repo_info_dict


def create_repository_admin_role(app: "ToolShedApp", repository: model.Repository):
    """
    Create a new role with name-spaced name based on the repository name and its owner's public user
    name.  This will ensure that the role name is unique.
    """
    sa_session = app.model.session
    name = get_repository_admin_role_name(str(repository.name), str(repository.user.username))
    description = "A user or group member with this role can administer this repository."
    role = model.Role(name=name, description=description, type=model.Role.types.SYSTEM)
    sa_session.add(role)
    # Associate the role with the repository owner.
    model.UserRoleAssociation(repository.user, role)
    # Associate the role with the repository.
    rra = model.RepositoryRoleAssociation(repository, role)
    sa_session.add(rra)
    return role


def create_repository(
    app: "ToolShedApp",
    name: str,
    type: str,
    description,
    long_description,
    user,
    category_ids: Optional[list[str]] = None,
    remote_repository_url=None,
    homepage_url=None,
) -> tuple[model.Repository, str]:
    """Create a new ToolShed repository"""
    category_ids = category_ids or []
    sa_session = app.model.session
    # Add the repository record to the database.
    repository = model.Repository(
        name=name,
        type=type,
        remote_repository_url=remote_repository_url,
        homepage_url=homepage_url,
        description=description,
        long_description=long_description,
        user=user,
    )
    sa_session.add(repository)
    if category_ids:
        # Create category associations
        for category_id in category_ids:
            category = sa_session.get(model.Category, app.security.decode_id(category_id))
            rca = model.RepositoryCategoryAssociation(repository, category)
            sa_session.add(rca)
    # Create an admin role for the repository.
    create_repository_admin_role(app, repository)
    # Create a temporary repo_path on disk.
    repository_path = tempfile.mkdtemp(
        dir=app.config.file_path,
        prefix=f"{repository.user.username}-{repository.name}",
    )
    # Created directory is readable, writable, and searchable only by the creating user ID,
    # but we need to make it world-readable so non-shed user can serve files (e.g. hgweb run as different user).
    os.chmod(repository_path, util.RWXR_XR_X)
    # Create the local repository.
    init_repository(repo_path=repository_path)
    # Create a .hg/hgrc file for the local repository.
    create_hgrc_file(app, repository, repo_path=repository_path)
    # Add an entry in the hgweb.config file for the local repository.
    lhs = f"{app.config.hgweb_repo_prefix}{repository.user.username}/{repository.name}"
    # Flush to get the id.
    session = sa_session()
    session.commit()
    final_repository_path = repository.ensure_hg_repository_path(app.config.file_path)
    os.rename(repository_path, final_repository_path)
    app.hgweb_config_manager.add_entry(lhs, final_repository_path)
    # Update the repository registry.
    app.repository_registry.add_entry(repository)
    message = f"Repository <b>{escape(str(repository.name))}</b> has been created."
    return repository, message


def generate_sharable_link_for_repository_in_tool_shed(
    repository: model.Repository, changeset_revision: Optional[str] = None
) -> str:
    """Generate the URL for sharing a repository that is in the tool shed."""
    base_url = web.url_for("/", qualified=True).rstrip("/")
    sharable_url = f"{base_url}/view/{repository.user.username}/{repository.name}"
    if changeset_revision:
        sharable_url += f"/{changeset_revision}"
    return sharable_url


def get_repository_in_tool_shed(app: "ToolShedApp", id, eagerload_columns=None):
    """Get a repository on the tool shed side from the database via id."""
    q = get_repository_query(app.model.context)
    if eagerload_columns:
        q = q.options(joinedload(*eagerload_columns))
    return q.get(app.security.decode_id(id))


def get_repo_info_dict(trans: "ProvidesRepositoriesContext", repository_id, changeset_revision):
    app = trans.app
    repository = get_repository_in_tool_shed(app, repository_id)
    repository_clone_url = generate_clone_url_for(trans, repository)
    repository_metadata = get_repository_metadata_by_changeset_revision(app, repository_id, changeset_revision)
    if not repository_metadata:
        # The received changeset_revision is no longer installable, so get the next changeset_revision
        # in the repository's changelog.  This generally occurs only with repositories of type
        # repository_suite_definition or tool_dependency_definition.
        next_downloadable_changeset_revision = get_next_downloadable_changeset_revision(
            app, repository, changeset_revision
        )
        if next_downloadable_changeset_revision and next_downloadable_changeset_revision != changeset_revision:
            repository_metadata = get_repository_metadata_by_changeset_revision(
                app, repository_id, next_downloadable_changeset_revision
            )
    if repository_metadata:
        # For now, we'll always assume that we'll get repository_metadata, but if we discover our assumption
        # is not valid we'll have to enhance the callers to handle repository_metadata values of None in the
        # returned repo_info_dict.
        metadata = repository_metadata.metadata
        if "tools" in metadata:
            includes_tools = True
        else:
            includes_tools = False
        includes_tools_for_display_in_tool_panel = repository_metadata.includes_tools_for_display_in_tool_panel
        repository_dependencies_dict = metadata.get("repository_dependencies", {})
        repository_dependencies = repository_dependencies_dict.get("repository_dependencies", [])
        (
            has_repository_dependencies,
            has_repository_dependencies_only_if_compiling_contained_td,
        ) = get_repository_dependency_types(repository_dependencies)
        if "tool_dependencies" in metadata:
            includes_tool_dependencies = True
        else:
            includes_tool_dependencies = False
    else:
        # Here's where we may have to handle enhancements to the callers. See above comment.
        includes_tools = False
        has_repository_dependencies = False
        has_repository_dependencies_only_if_compiling_contained_td = False
        includes_tool_dependencies = False
        includes_tools_for_display_in_tool_panel = False
    repo_path = repository.repo_path(app)
    ctx_rev = str(changeset2rev(repo_path, changeset_revision))
    repo_info_dict = create_repo_info_dict(
        app=app,
        repository_clone_url=repository_clone_url,
        changeset_revision=changeset_revision,
        ctx_rev=ctx_rev,
        repository_owner=repository.user.username,
        repository_name=repository.name,
        repository=repository,
        repository_metadata=repository_metadata,
        tool_dependencies=None,
        repository_dependencies=None,
        trans=trans,
    )
    return (
        repo_info_dict,
        includes_tools,
        includes_tool_dependencies,
        includes_tools_for_display_in_tool_panel,
        has_repository_dependencies,
        has_repository_dependencies_only_if_compiling_contained_td,
    )


def get_repositories_by_category(
    app: "ToolShedApp",
    category_id,
    installable: bool = False,
    sort_order="asc",
    sort_key="name",
    page: Optional[int] = None,
    per_page: int = 25,
):
    repositories = []
    for repository in get_repositories(
        app.model.session,
        category_id,
        installable,
        sort_order,
        sort_key,
        page,
        per_page,
    ):
        default_value_mapper = {
            "id": app.security.encode_id,
            "user_id": app.security.encode_id,
            "repository_id": app.security.encode_id,
        }
        repository_dict = repository.to_dict(value_mapper=default_value_mapper)
        repository_dict["metadata"] = {}
        for changeset, changehash in repository.installable_revisions(app):
            metadata = repository_metadata_by_changeset_revision(app.model, repository.id, changehash)
            assert metadata
            repository_dict["metadata"][f"{changeset}:{changehash}"] = metadata.to_dict(
                value_mapper=default_value_mapper
            )
        if installable:
            if len(repository.installable_revisions(app)):
                repositories.append(repository_dict)
        else:
            repositories.append(repository_dict)
    return repositories


def handle_role_associations(app: "ToolShedApp", role, repository, **kwd):
    sa_session = app.model.session
    message = escape(kwd.get("message", ""))
    status = kwd.get("status", "done")
    repository_owner = repository.user
    if kwd.get("manage_role_associations_button", False):
        in_users_list = util.listify(kwd.get("in_users", []))
        users = [y for y in (sa_session.get(model.User, x) for x in in_users_list) if y is not None]
        # Make sure the repository owner is always associated with the repostory's admin role.
        owner_associated = False
        for user in users:
            if user.id == repository_owner.id:
                owner_associated = True
                break
        if not owner_associated:
            users.append(repository_owner)
            message += "The repository owner must always be associated with the repository's administrator role.  "
            status = "error"
        in_groups_list = util.listify(kwd.get("in_groups", []))
        groups = [sa_session.get(model.Group, x) for x in in_groups_list]
        in_repositories = [repository]
        app.security_agent.set_entity_role_associations(
            roles=[role], users=users, groups=groups, repositories=in_repositories
        )
        sa_session.refresh(role)
        message += f"Role <b>{escape(str(role.name))}</b> has been associated with {len(users)} users, {len(groups)} groups and {len(in_repositories)} repositories.  "
    in_users = []
    out_users = []
    in_groups = []
    out_groups = []
    for user in get_current_users(sa_session):
        if user in [x.user for x in role.users]:
            in_users.append((user.id, user.email))
        else:
            out_users.append((user.id, user.email))
    for group in get_current_groups(sa_session):
        if group in [x.group for x in role.groups]:
            in_groups.append((group.id, group.name))
        else:
            out_groups.append((group.id, group.name))
    associations_dict = dict(
        in_users=in_users,
        out_users=out_users,
        in_groups=in_groups,
        out_groups=out_groups,
        message=message,
        status=status,
    )
    return associations_dict


def change_repository_name_in_hgrc_file(hgrc_file: str, new_name: str) -> None:
    config = configparser.ConfigParser()
    config.read(hgrc_file)
    config.set("web", "name", new_name)
    with open(hgrc_file, "w") as fh:
        config.write(fh)


def update_repository(
    trans: "ProvidesUserContext", id: str, **kwds
) -> tuple[Optional[model.Repository], Optional[str]]:
    """Update an existing ToolShed repository"""
    app = trans.app
    sa_session = app.model.session
    repository = sa_session.get(model.Repository, app.security.decode_id(id))
    if repository is None:
        return None, "Unknown repository ID"

    if not (trans.user_is_admin or app.security_agent.user_can_administer_repository(trans.user, repository)):
        message = "You are not the owner of this repository, so you cannot administer it."
        return None, message

    return update_validated_repository(trans, repository, **kwds)


def update_validated_repository(
    trans: "ProvidesUserContext", repository: model.Repository, **kwds
) -> tuple[Optional[model.Repository], Optional[str]]:
    """Update an existing ToolShed repository metadata once permissions have been checked."""
    app = trans.app
    sa_session = app.model.session
    message = None
    flush_needed = False

    # Allowlist properties that can be changed via this method
    for key in ("type", "description", "long_description", "remote_repository_url", "homepage_url"):
        # If that key is available, not None and different than what's in the model
        if key in kwds and kwds[key] is not None and kwds[key] != getattr(repository, key):
            setattr(repository, key, kwds[key])
            flush_needed = True

    if "category_ids" in kwds and isinstance(kwds["category_ids"], list):

        # Remove existing category associations
        delete_repository_category_associations(sa_session, model.RepositoryCategoryAssociation, repository.id)

        # Then (re)create category associations
        for category_id in kwds["category_ids"]:
            category = sa_session.get(model.Category, app.security.decode_id(category_id))
            if category:
                rca = model.RepositoryCategoryAssociation(repository, category)
                sa_session.add(rca)
            else:
                pass
        flush_needed = True

    # However some properties are special, like 'name'
    if "name" in kwds and kwds["name"] is not None and repository.name != kwds["name"]:
        if repository.times_downloaded != 0:
            message = "Repository names cannot be changed if the repository has been cloned."
        else:
            message = validate_repository_name(trans.app, kwds["name"], trans.user)
        if message:
            return None, message

        repo_dir = repository.repo_path(app)
        # Change the entry in the hgweb.config file for the repository.
        old_lhs = f"{trans.app.config.hgweb_repo_prefix}{repository.user.username}/{repository.name}"
        new_lhs = f"{trans.app.config.hgweb_repo_prefix}{repository.user.username}/{kwds['name']}"
        trans.app.hgweb_config_manager.change_entry(old_lhs, new_lhs, repo_dir)

        # Change the entry in the repository's hgrc file.
        hgrc_file = get_hgrc_path(repo_dir)
        change_repository_name_in_hgrc_file(hgrc_file, kwds["name"])

        # Rename the repository's admin role to match the new repository name.
        repository_admin_role = repository.admin_role
        repository_admin_role.name = get_repository_admin_role_name(str(kwds["name"]), str(repository.user.username))
        trans.sa_session.add(repository_admin_role)
        repository.name = kwds["name"]
        flush_needed = True

    if flush_needed:
        trans.sa_session.add(repository)
        trans.sa_session.commit()
        message = "The repository information has been updated."
    else:
        message = None
    return repository, message


def validate_repository_name(app: "ToolShedApp", name, user):
    """
    Validate whether the given name qualifies as a new TS repo name.
    Repository names must be unique for each user, must be at least two characters
    in length and must contain only lower-case letters, numbers, and the '_' character.
    """
    if name in ["None", None, ""]:
        return "Enter the required repository name."
    if name in ["repos"]:
        return f"The term '{name}' is a reserved word in the Tool Shed, so it cannot be used as a repository name."
    check_existing = get_repository_by_name_and_owner(app.model.context, name, user.username)
    if check_existing is not None:
        if check_existing.deleted:
            return f"You own a deleted repository named <b>{escape(name)}</b>, please choose a different name."
        else:
            return f"You already own a repository named <b>{escape(name)}</b>, please choose a different name."
    if len(name) < 2:
        return "Repository names must be at least 2 characters in length."
    if len(name) > 80:
        return "Repository names cannot be more than 80 characters in length."
    if not (VALID_REPOSITORYNAME_RE.match(name)):
        return "Repository names must contain only lower-case letters, numbers and underscore."
    return ""


def get_repositories(
    session: "scoped_session",
    category_id,
    installable: bool,
    sort_order,
    sort_key,
    page: Optional[int],
    per_page: int,
):
    stmt = (
        select(model.Repository)
        .join(
            model.RepositoryCategoryAssociation,
            model.Repository.id == model.RepositoryCategoryAssociation.repository_id,
        )
        .join(model.User, model.User.id == model.Repository.user_id)
        .where(model.RepositoryCategoryAssociation.category_id == category_id)
    )
    if installable:
        stmt1 = select(model.RepositoryMetadata.repository_id)
        stmt = stmt.where(model.Repository.id.in_(stmt1))

    if sort_key == "owner":
        sort_col = model.User.username
    else:
        sort_col = model.Repository.name
    sort_by = sort_col.desc() if sort_order == "desc" else sort_col
    stmt = stmt.order_by(sort_by)

    if page is not None:
        page = int(page)
        stmt = stmt.limit(per_page)
        if page > 1:
            stmt = stmt.offset((page - 1) * per_page)

    return session.scalars(stmt)


def get_current_users(session: "scoped_session"):
    stmt = select(model.User).where(model.User.deleted == false()).order_by(model.User.email)
    return session.scalars(stmt)


def get_current_groups(session: "scoped_session"):
    stmt = select(model.Group).where(model.Group.deleted == false()).order_by(model.Group.name)
    return session.scalars(stmt)


def delete_repository_category_associations(session, repository_category_assoc_model, repository_id):
    stmt = delete(repository_category_assoc_model).where(repository_category_assoc_model.repository_id == repository_id)
    return session.execute(stmt)


__all__ = (
    "change_repository_name_in_hgrc_file",
    "create_or_update_tool_shed_repository",
    "create_repo_info_dict",
    "create_repository_admin_role",
    "create_repository",
    "extract_components_from_tuple",
    "generate_sharable_link_for_repository_in_tool_shed",
    "generate_tool_shed_repository_install_dir",
    "get_absolute_path_to_file_in_repository",
    "get_installed_repository",
    "get_installed_tool_shed_repository",
    "get_prior_import_or_install_required_dict",
    "get_repo_info_dict",
    "get_repo_info_tuple_contents",
    "get_repositories_by_category",
    "get_repository_admin_role_name",
    "get_repository_and_repository_dependencies_from_repo_info_dict",
    "get_repository_by_id",
    "get_repository_dependency_types",
    "get_repository_for_dependency_relationship",
    "get_repository_ids_requiring_prior_import_or_install",
    "get_repository_in_tool_shed",
    "get_repository_owner",
    "get_repository_owner_from_clone_url",
    "get_role_by_id",
    "get_tool_shed_from_clone_url",
    "get_tool_shed_repository_by_id",
    "get_tool_shed_status_for_installed_repository",
    "handle_role_associations",
    "is_tool_shed_client",
    "repository_was_previously_installed",
    "set_repository_attributes",
    "update_repository",
    "validate_repository_name",
)



admin_user = None
admin_user_private_role = None
admin_email = "test@bx.psu.edu"
admin_username = "admin-user"

test_user_1 = None
test_user_1_private_role = None
test_user_1_email = "test-1@bx.psu.edu"
test_user_1_name = "user1"

test_user_2 = None
test_user_2_private_role = None
test_user_2_email = "test-2@bx.psu.edu"
test_user_2_name = "user2"

test_user_3 = None
test_user_3_private_role = None
test_user_3_email = "test-3@bx.psu.edu"
test_user_3_name = "user3"

complex_repository_dependency_template = """<?xml version="1.0"?>
<tool_dependency>
    <package name="${package}" version="${version}">
${dependency_lines}
    </package>
</tool_dependency>
"""

new_repository_dependencies_xml = """<?xml version="1.0"?>
<repositories${description}>
${dependency_lines}
</repositories>
"""

new_repository_dependencies_line = """    <repository toolshed="${toolshed_url}" name="${repository_name}" owner="${owner}" changeset_revision="${changeset_revision}"${prior_installation_required} />"""



import logging

import tool_shed.repository_types.util as rt_util
from tool_shed.repository_types.metadata import Metadata

log = logging.getLogger(__name__)


class Unrestricted(Metadata):
    def __init__(self):
        self.type = rt_util.UNRESTRICTED
        self.label = "Unrestricted"

    def is_valid_for_type(self, repository, revisions_to_check=None):
        """A repository's type can only be changed to the unrestricted type if it is new or has never been installed."""
        if repository.is_new():
            return True
        if repository.times_downloaded == 0:
            return True
        return False



from typing import (
    Optional,
)

from fastapi import Body

from tool_shed.context import SessionRequestContext
from tool_shed.managers.categories import CategoryManager
from tool_shed.managers.repositories import repositories_by_category
from tool_shed_client.schema import (
    Category as CategoryResponse,
    CreateCategoryRequest,
    RepositoriesByCategory,
)
from . import (
    CategoryIdPathParam,
    CategoryRepositoriesInstallableQueryParam,
    CategoryRepositoriesPageQueryParam,
    CategoryRepositoriesSortKeyQueryParam,
    CategoryRepositoriesSortOrderQueryParam,
    depends,
    DependsOnTrans,
    Router,
)

router = Router(tags=["categories"])


@router.cbv
class FastAPICategories:
    category_manager: CategoryManager = depends(CategoryManager)

    @router.post(
        "/api/categories",
        description="create a category",
        operation_id="categories__create",
        require_admin=True,
    )
    def create(
        self, trans: SessionRequestContext = DependsOnTrans, request: CreateCategoryRequest = Body(...)
    ) -> CategoryResponse:
        category = self.category_manager.create(trans, request)
        return self.category_manager.to_model(category)

    @router.get(
        "/api/categories",
        description="index category",
        operation_id="categories__index",
    )
    def index(self, trans: SessionRequestContext = DependsOnTrans) -> list[CategoryResponse]:
        """
        Return a list of dictionaries that contain information about each Category.
        """
        deleted = False
        categories = self.category_manager.index_db(trans, deleted)
        return [self.category_manager.to_model(c) for c in categories]

    @router.get(
        "/api/categories/{encoded_category_id}",
        description="show category",
        operation_id="categories__show",
    )
    def show(self, encoded_category_id: str = CategoryIdPathParam) -> CategoryResponse:
        """
        Return a list of dictionaries that contain information about each Category.
        """
        category = self.category_manager.get(encoded_category_id)
        return self.category_manager.to_model(category)

    @router.get(
        "/api/categories/{encoded_category_id}/repositories",
        description="display repositories by category",
        operation_id="categories__repositories",
    )
    def repositories(
        self,
        trans: SessionRequestContext = DependsOnTrans,
        encoded_category_id: str = CategoryIdPathParam,
        installable: bool = CategoryRepositoriesInstallableQueryParam,
        sort_key: str = CategoryRepositoriesSortKeyQueryParam,
        sort_order: str = CategoryRepositoriesSortOrderQueryParam,
        page: Optional[int] = CategoryRepositoriesPageQueryParam,
    ) -> RepositoriesByCategory:
        return repositories_by_category(
            trans.app,
            encoded_category_id,
            page=page,
            sort_key=sort_key,
            sort_order=sort_order,
            installable=installable,
        )



from ..base.api import ShedApiTestCase


class TestShedConfigurationApi(ShedApiTestCase):
    def test_version(self) -> None:
        version = self.populator.version()
        assert version.version
        assert version.version_major



from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

column_maker_repository_name = "column_maker_0030"
column_maker_repository_description = "Add column"
column_maker_repository_long_description = "Compute an expression on every row"

emboss_repository_name = "emboss_0030"
emboss_5_repository_name = "emboss_5_0030"
emboss_6_repository_name = "emboss_6_0030"
emboss_repository_description = "Galaxy wrappers for Emboss version 5.0.0 tools"
emboss_repository_long_description = "Galaxy wrappers for Emboss version 5.0.0 tools"

running_standalone = False


class TestUninstallingAndReinstallingRepositories(ShedTwillTestCase):
    """Test uninstalling and reinstalling a repository with repository dependency revisions."""

    requires_galaxy = True

    def test_0000_initiate_users(self):
        """Create necessary user accounts."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_ensure_repositories_and_categories_exist(self):
        """Create the 0030 category and upload the emboss repository to the tool shed, if necessary."""
        global running_standalone
        category = self.create_category(
            name="Test 0030 Repository Dependency Revisions",
            description="Tests for a repository with tool dependencies.",
        )
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        column_maker_repository = self.get_or_create_repository(
            name=column_maker_repository_name,
            description=column_maker_repository_description,
            long_description=column_maker_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        if self.repository_is_new(column_maker_repository):
            running_standalone = True
            emboss_5_repository = self.get_or_create_repository(
                name=emboss_5_repository_name,
                description=emboss_repository_description,
                long_description=emboss_repository_long_description,
                owner=common.test_user_1_name,
                category=category,
                strings_displayed=[],
            )
            self.commit_tar_to_repository(
                column_maker_repository,
                "column_maker/column_maker.tar",
                commit_message="Uploaded column maker tarball.",
            )
            repository_dependencies_path = self.generate_temp_path("test_1030", additional_paths=["emboss", "5"])
            column_maker_tuple = (
                self.url,
                column_maker_repository.name,
                column_maker_repository.owner,
                self.get_repository_tip(column_maker_repository),
            )
            self.create_repository_dependency(
                repository=emboss_5_repository,
                repository_tuples=[column_maker_tuple],
                filepath=repository_dependencies_path,
            )
            emboss_6_repository = self.get_or_create_repository(
                name=emboss_6_repository_name,
                description=emboss_repository_description,
                long_description=emboss_repository_long_description,
                owner=common.test_user_1_name,
                category=category,
                strings_displayed=[],
            )
            self.commit_tar_to_repository(
                emboss_6_repository,
                "emboss/emboss.tar",
                commit_message="Uploaded tool tarball.",
            )
            repository_dependencies_path = self.generate_temp_path("test_1030", additional_paths=["emboss", "6"])
            column_maker_tuple = (
                self.url,
                column_maker_repository.name,
                column_maker_repository.owner,
                self.get_repository_tip(column_maker_repository),
            )
            self.create_repository_dependency(
                repository=emboss_6_repository,
                repository_tuples=[column_maker_tuple],
                filepath=repository_dependencies_path,
            )
            emboss_repository = self.get_or_create_repository(
                name=emboss_repository_name,
                description=emboss_repository_description,
                long_description=emboss_repository_long_description,
                owner=common.test_user_1_name,
                category=category,
                strings_displayed=[],
            )
            self.commit_tar_to_repository(
                emboss_repository,
                "emboss/emboss.tar",
                commit_message="Uploaded tool tarball.",
            )
            repository_dependencies_path = self.generate_temp_path("test_1030", additional_paths=["emboss", "5"])
            dependency_tuple = (
                self.url,
                emboss_5_repository.name,
                emboss_5_repository.owner,
                self.get_repository_tip(emboss_5_repository),
            )
            self.create_repository_dependency(
                repository=emboss_repository,
                repository_tuples=[dependency_tuple],
                filepath=repository_dependencies_path,
            )
            dependency_tuple = (
                self.url,
                emboss_6_repository.name,
                emboss_6_repository.owner,
                self.get_repository_tip(emboss_6_repository),
            )
            self.create_repository_dependency(
                repository=emboss_repository,
                repository_tuples=[dependency_tuple],
                filepath=repository_dependencies_path,
            )

    def test_0010_install_emboss_repository(self):
        """Install the emboss repository into the Galaxy instance."""
        self._install_repository(
            emboss_repository_name,
            common.test_user_1_name,
            "Test 0030 Repository Dependency Revisions",
            new_tool_panel_section_label="test_1210",
        )
        self._assert_has_installed_repos_with_names(emboss_repository_name)

    def test_0015_uninstall_emboss_repository(self):
        """Uninstall the emboss repository."""
        installed_repository = self._get_installed_repository_by_name_owner(
            emboss_repository_name, common.test_user_1_name
        )
        self._uninstall_repository(installed_repository)
        self._assert_has_no_installed_repos_with_names(emboss_repository_name)

    def test_0020_reinstall_emboss_repository(self):
        """Reinstall the emboss repository."""
        installed_repository = self._get_installed_repository_by_name_owner(
            emboss_repository_name, common.test_user_1_name
        )
        self.reinstall_repository_api(installed_repository)
        self._assert_has_installed_repos_with_names(emboss_repository_name)
        self._assert_has_valid_tool_with_name("emboss")

    def test_0025_deactivate_emboss_repository(self):
        """Deactivate the emboss repository without removing it from disk."""
        installed_repository = self._get_installed_repository_by_name_owner(
            emboss_repository_name, common.test_user_1_name
        )
        self.deactivate_repository(installed_repository)
        self._assert_has_no_installed_repos_with_names(emboss_repository_name)

    def test_0030_reactivate_emboss_repository(self):
        """Reactivate the emboss repository and verify that it now shows up in the list of installed repositories."""
        installed_repository = self._get_installed_repository_by_name_owner(
            emboss_repository_name, common.test_user_1_name
        )
        self.reactivate_repository(installed_repository)
        self._assert_has_installed_repos_with_names(emboss_repository_name)
        self._assert_has_valid_tool_with_name("emboss")
        self._assert_repo_has_tool_with_id(installed_repository, "EMBOSS: antigenic1")



from fastapi import Request

from galaxy.webapps.galaxy.services.authenticate import (
    APIKeyResponse,
    AuthenticationService,
)
from . import (
    depends,
    Router,
)

router = Router(tags=["authenticate"])


@router.cbv
class FastAPIAuthenticate:
    authentication_service: AuthenticationService = depends(AuthenticationService)

    @router.get(
        "/api/authenticate/baseauth",
        summary="Returns returns an API key for authenticated user based on BaseAuth headers.",
        operation_id="authenticate__baseauth",
    )
    def get_api_key(self, request: Request) -> APIKeyResponse:
        authorization = request.headers.get("Authorization")
        auth = {"HTTP_AUTHORIZATION": authorization}
        return self.authentication_service.get_api_key(auth, request)



"""
API key retrieval through BaseAuth
Sample usage:

.. code-block::

    curl --user zipzap@foo.com:password http://localhost:9009/api/authenticate/baseauth

Returns

.. code-block:: json

    {
        "api_key": "<some api key>"
    }

"""

import logging

from galaxy.web import expose_api_anonymous_and_sessionless
from galaxy.webapps.galaxy.api import depends
from galaxy.webapps.galaxy.services.authenticate import AuthenticationService
from . import BaseShedAPIController

log = logging.getLogger(__name__)


class ToolShedAuthenticationController(BaseShedAPIController):
    authentication_service = depends(AuthenticationService)

    @expose_api_anonymous_and_sessionless
    def get_tool_shed_api_key(self, trans, **kwd):
        """
        GET /api/authenticate/baseauth

        returns an API key for authenticated user based on BaseAuth headers

        :returns: api_key in json format
        :rtype:   dict

        :raises: ObjectNotFound, HTTPBadRequest
        """
        return self.authentication_service.get_api_key(trans.environ, trans.request)



import json
import logging
import os
import socket
import string
from typing import (
    TYPE_CHECKING,
)

from sqlalchemy import (
    false,
    func,
    select,
    true,
)
from sqlalchemy.orm import scoped_session

from galaxy import util
from galaxy.tool_shed.util.shed_util_common import get_user
from galaxy.util import (
    checkers,
    unicodify,
)
from galaxy.web import url_for
from tool_shed.util import (
    basic_util,
    common_util,
    hg_util,
    repository_util,
)
from tool_shed.webapp import model

if TYPE_CHECKING:
    from tool_shed.structured_app import ToolShedApp
    from tool_shed.webapp.model import Repository

log = logging.getLogger(__name__)

MAX_CONTENT_SIZE = 1048576
DATATYPES_CONFIG_FILENAME = "datatypes_conf.xml"
REPOSITORY_DATA_MANAGER_CONFIG_FILENAME = "data_manager_conf.xml"

new_repo_email_alert_template = """
Sharable link:         ${sharable_link}
Repository name:       ${repository_name}
Revision:              ${revision}
Change description:
${description}

Uploaded by:           ${username}
Date content uploaded: ${display_date}

${content_alert_str}

-----------------------------------------------------------------------------
This change alert was sent from the Galaxy tool shed hosted on the server
"${host}"
-----------------------------------------------------------------------------
You received this alert because you registered to receive email when
new repositories were created in the Galaxy tool shed named "${host}".
-----------------------------------------------------------------------------
"""

email_alert_template = """
Sharable link:         ${sharable_link}
Repository name:       ${repository_name}
Revision:              ${revision}
Change description:
${description}

Changed by:     ${username}
Date of change: ${display_date}

${content_alert_str}

-----------------------------------------------------------------------------
This change alert was sent from the Galaxy tool shed hosted on the server
"${host}"
-----------------------------------------------------------------------------
You received this alert because you registered to receive email whenever
changes were made to the repository named "${repository_name}".
-----------------------------------------------------------------------------
"""


def count_repositories_in_category(app: "ToolShedApp", category_id: str) -> int:
    stmt = (
        select(func.count())
        .select_from(model.RepositoryCategoryAssociation)
        .where(model.RepositoryCategoryAssociation.category_id == app.security.decode_id(category_id))
    )
    count = app.model.session.scalar(stmt)
    assert count is not None
    return count


def get_categories(app: "ToolShedApp"):
    """Get all categories from the database."""
    sa_session = app.model.session
    stmt = select(model.Category).where(model.Category.deleted == false()).order_by(model.Category.name)
    return sa_session.scalars(stmt).all()


def get_category(app: "ToolShedApp", id: str):
    """Get a category from the database."""
    sa_session = app.model.session
    return sa_session.get(model.Category, app.security.decode_id(id))


def get_category_by_name(app: "ToolShedApp", name: str):
    """Get a category from the database via name."""
    sa_session = app.model.session
    stmt = select(model.Category).filter_by(name=name).limit(1)
    return sa_session.scalars(stmt).first()


def get_repository_file_contents(app: "ToolShedApp", file_path: str, repository_id: str, is_admin: bool = False) -> str:
    """Return the display-safe contents of a repository file for display in a browser."""
    safe_str = ""
    if not _is_path_browsable(app, file_path, repository_id, is_admin):
        log.warning("Request tries to access a file outside of the repository location. File path: %s", file_path)
        return "Invalid file path"
    # Symlink targets are checked by _is_path_browsable
    if os.path.islink(file_path):
        safe_str = f"link to: {basic_util.to_html_string(os.readlink(file_path))}"
        return safe_str
    elif checkers.is_gzip(file_path):
        return "<br/>gzip compressed file<br/>"
    elif checkers.is_bz2(file_path):
        return "<br/>bz2 compressed file<br/>"
    elif checkers.is_zip(file_path):
        return "<br/>zip compressed file<br/>"
    elif checkers.check_binary(file_path):
        return "<br/>Binary file<br/>"
    else:
        with open(file_path) as fh:
            for line in fh:
                safe_str = f"{safe_str}{basic_util.to_html_string(line)}"
                # Stop reading after string is larger than MAX_CONTENT_SIZE.
                if len(safe_str) > MAX_CONTENT_SIZE:
                    large_str = f"<br/>File contents truncated because file size is larger than maximum viewing size of {util.nice_size(MAX_CONTENT_SIZE)}<br/>"
                    safe_str = f"{safe_str}{large_str}"
                    break

        if len(safe_str) > basic_util.MAX_DISPLAY_SIZE:
            # Eliminate the middle of the file to display a file no larger than basic_util.MAX_DISPLAY_SIZE.
            # This may not be ideal if the file is larger than MAX_CONTENT_SIZE.
            join_by_str = f"<br/><br/>...some text eliminated here because file size is larger than maximum viewing size of {util.nice_size(basic_util.MAX_DISPLAY_SIZE)}...<br/><br/>"
            safe_str = util.shrink_string_by_size(
                safe_str,
                basic_util.MAX_DISPLAY_SIZE,
                join_by=join_by_str,
                left_larger=True,
                beginning_on_size_error=True,
            )
        return safe_str


def get_repository_files(folder_path):
    """Return the file hierarchy of a tool shed repository."""
    contents = []
    for item in os.listdir(folder_path):
        # Skip .hg directories
        if item.startswith(".hg"):
            continue
        contents.append(item)
    if contents:
        contents.sort()
    return contents


def get_repository_from_refresh_on_change(app: "ToolShedApp", **kwd):
    # The changeset_revision_select_field in several grids performs a refresh_on_change which sends in request parameters like
    # changeset_revison_1, changeset_revision_2, etc.  One of the many select fields on the grid performed the refresh_on_change,
    # so we loop through all of the received values to see which value is not the repository tip.  If we find it, we know the
    # refresh_on_change occurred and we have the necessary repository id and change set revision to pass on.
    repository_id = None
    v = None
    for k, v in kwd.items():
        changeset_revision_str = "changeset_revision_"
        if k.startswith(changeset_revision_str):
            repository_id = app.security.encode_id(int(k.lstrip(changeset_revision_str)))
            repository = repository_util.get_repository_in_tool_shed(app, repository_id)
            if repository.tip() != v:
                return v, repository
    # This should never be reached - raise an exception?
    return v, None


def handle_email_alerts(
    app: "ToolShedApp",
    host: str,
    repository: "Repository",
    content_alert_str: str = "",
    new_repo_alert: bool = False,
    admin_only: bool = False,
) -> None:
    """
    There are 2 complementary features that enable a tool shed user to receive email notification:

    1. Within User Preferences, they can elect to receive email when the first (or first valid)
       change set is produced for a new repository.
    2. When viewing or managing a repository, they can check the box labeled "Receive email alerts"
       which caused them to receive email alerts when updates to the repository occur.  This same feature
       is available on a per-repository basis on the repository grid within the tool shed.

    There are currently 4 scenarios for sending email notification when a change is made to a repository:

    1. An admin user elects to receive email when the first change set is produced for a new repository
       from User Preferences.  The change set does not have to include any valid content.  This allows for
       the capture of inappropriate content being uploaded to new repositories.
    2. A regular user elects to receive email when the first valid change set is produced for a new repository
       from User Preferences.  This differs from 1 above in that the user will not receive email until a
       change set that includes valid content is produced.
    3. An admin user checks the "Receive email alerts" check box on the manage repository page.  Since the
       user is an admin user, the email will include information about both HTML and image content that was
       included in the change set.
    4. A regular user checks the "Receive email alerts" check box on the manage repository page.  Since the
       user is not an admin user, the email will not include any information about both HTML and image content
       that was included in the change set.

    """
    sa_session = app.model.session
    repo = repository.hg_repo
    sharable_link = repository_util.generate_sharable_link_for_repository_in_tool_shed(
        repository, changeset_revision=None
    )
    smtp_server = app.config.smtp_server
    if smtp_server and (new_repo_alert or repository.email_alerts):
        # Send email alert to users that want them.
        if app.config.email_from is not None:
            email_from = app.config.email_from
        elif host.split(":")[0] in ["localhost", "127.0.0.1", "0.0.0.0"]:
            email_from = f"galaxy-no-reply@{socket.getfqdn()}"
        else:
            email_from = f"galaxy-no-reply@{host.split(':')[0]}"
        ctx = repo[repo.changelog.tip()]
        username = unicodify(ctx.user())
        try:
            username = username.split()[0]
        except Exception:
            pass
        # We'll use 2 template bodies because we only want to send content
        # alerts to tool shed admin users.
        if new_repo_alert:
            template = new_repo_email_alert_template
        else:
            template = email_alert_template
        display_date = hg_util.get_readable_ctx_date(ctx)
        description = unicodify(ctx.description())
        revision = f"{ctx.rev()}:{ctx}"
        admin_body = string.Template(template).safe_substitute(
            host=host,
            sharable_link=sharable_link,
            repository_name=repository.name,
            revision=revision,
            display_date=display_date,
            description=description,
            username=username,
            content_alert_str=content_alert_str,
        )
        body = string.Template(template).safe_substitute(
            host=host,
            sharable_link=sharable_link,
            repository_name=repository.name,
            revision=revision,
            display_date=display_date,
            description=description,
            username=username,
            content_alert_str="",
        )
        admin_users = app.config.get("admin_users", "").split(",")
        frm = email_from
        if new_repo_alert:
            subject = f"Galaxy tool shed alert for new repository named {str(repository.name)}"
            subject = subject[:80]
            email_alerts = []
            for user in get_users_with_repo_alert(sa_session, model.User):
                if admin_only:
                    if user.email in admin_users:
                        email_alerts.append(user.email)
                else:
                    email_alerts.append(user.email)
        else:
            subject = f"Galaxy tool shed update alert for repository named {str(repository.name)}"
            email_alerts = json.loads(repository.email_alerts)  # type:ignore[arg-type]
        for email in email_alerts:
            to = email.strip()
            # Send it
            try:
                if to in admin_users:
                    util.send_mail(frm, to, subject, admin_body, app.config)
                else:
                    util.send_mail(frm, to, subject, body, app.config)
            except Exception:
                log.exception("An error occurred sending a tool shed repository update alert by email.")


def _is_path_browsable(app: "ToolShedApp", path: str, repository_id: str, is_admin: bool = False) -> bool:
    """
    Detects whether the given path is browsable i.e. is within the
    allowed repository folders. Admins can additionaly browse folders
    with tool dependencies.
    """
    if is_admin and is_path_within_dependency_dir(app, path):
        return True
    return is_path_within_repo(app, path, repository_id)


def is_path_within_dependency_dir(app: "ToolShedApp", path: str) -> bool:
    """
    Detect whether the given path is within the tool_dependency_dir folder on the disk.
    (Specified by the config option). Use to filter malicious symlinks targeting outside paths.
    """
    allowed = False
    resolved_path = os.path.realpath(path)
    if tool_dependency_dir := app.config.get("tool_dependency_dir", None):
        dependency_path = os.path.abspath(tool_dependency_dir)
        allowed = os.path.commonprefix([dependency_path, resolved_path]) == dependency_path
    return allowed


def is_path_within_repo(app: "ToolShedApp", path: str, repository_id: str) -> bool:
    """
    Detect whether the given path is within the repository folder on the disk.
    Use to filter malicious symlinks targeting outside paths.
    """
    repo_path = os.path.abspath(repository_util.get_repository_by_id(app, repository_id).repo_path(app))
    resolved_path = os.path.realpath(path)
    return os.path.commonprefix([repo_path, resolved_path]) == repo_path


def open_repository_files_folder(
    app: "ToolShedApp", folder_path: str, repository_id: str, is_admin: bool = False
) -> list:
    """
    Return a list of dictionaries, each of which contains information for a file or directory contained
    within a directory in a repository file hierarchy.
    """
    if not _is_path_browsable(app, folder_path, repository_id, is_admin):
        log.warning("Request tries to access a folder outside of the allowed locations. Folder path: %s", folder_path)
        return []
    try:
        files_list = get_repository_files(folder_path)
    except OSError as e:
        if str(e).find("No such file or directory") >= 0:
            # We have a repository with no contents.
            return []
    folder_contents = []
    for filename in files_list:
        is_folder = False
        full_path = os.path.join(folder_path, filename)
        is_link = os.path.islink(full_path)
        path_is_browsable = _is_path_browsable(app, full_path, repository_id)
        if is_link and not path_is_browsable:
            log.warning(
                f"Valid folder contains a symlink outside of the repository location. Link found in: {str(full_path)}"
            )
        if filename:
            if os.path.isdir(full_path) and path_is_browsable:
                # Append a '/' character so that our jquery dynatree will function properly.
                filename = f"{filename}/"
                full_path = f"{full_path}/"
                is_folder = True
            node = {
                "title": filename,
                "isFolder": is_folder,
                "isLazy": is_folder,
                "tooltip": full_path,
                "key": full_path,
            }
            folder_contents.append(node)
    return folder_contents


def tool_shed_is_this_tool_shed(toolshed_base_url, trans=None):
    """Determine if a tool shed is the current tool shed."""
    cleaned_toolshed_base_url = common_util.remove_protocol_from_tool_shed_url(toolshed_base_url)
    hostname = trans.repositories_hostname if trans else str(url_for("/", qualified=True))
    cleaned_tool_shed = common_util.remove_protocol_from_tool_shed_url(hostname)
    return cleaned_toolshed_base_url == cleaned_tool_shed


def get_users_with_repo_alert(session: scoped_session, user_model):
    stmt = select(user_model).where(user_model.deleted == false()).where(user_model.new_repo_alert == true())
    return session.scalars(stmt)


__all__ = (
    "count_repositories_in_category",
    "get_categories",
    "get_category",
    "get_category_by_name",
    "get_repository_file_contents",
    "get_user",
    "handle_email_alerts",
    "is_path_within_dependency_dir",
    "is_path_within_repo",
    "open_repository_files_folder",
    "tool_shed_is_this_tool_shed",
)



from tool_shed.structured_app import ToolShedApp
from tool_shed_client.schema import Version
from . import (
    depends,
    Router,
)

router = Router(tags=["configuration"])


@router.cbv
class FastAPIConfiguration:
    app: ToolShedApp = depends(ToolShedApp)

    @router.get(
        "/api/version",
        operation_id="configuration__version",
    )
    def version(self) -> Version:
        return Version(
            version_major=self.app.config.version_major,
            version=self.app.config.version,
            api_version="v2",
        )



from typing import (
    Any,
    cast,
    Optional,
)

from starlette.datastructures import URL

from galaxy.exceptions import ObjectNotFound
from galaxy.util.tool_shed.common_util import remove_protocol_and_user_from_clone_url
from galaxy.version import VERSION
from tool_shed.context import ProvidesRepositoriesContext
from tool_shed.structured_app import ToolShedApp
from tool_shed.util.metadata_util import get_current_repository_metadata_for_changeset_revision
from tool_shed.webapp.model import (
    Repository,
    RepositoryMetadata,
)
from tool_shed_client.schema.trs import (
    DescriptorType,
    Tool,
    ToolClass,
    ToolVersion,
)
from tool_shed_client.schema.trs_service_info import (
    Organization,
    Service,
    ServiceType,
)
from tool_shed_client.trs_util import decode_identifier
from .repositories import guid_to_repository

TRS_SERVICE_NAME = "Tool Shed TRS API"
TRS_SERVICE_DESCRIPTION = "Serves tool shed repository tools according to the GA4GH TRS specification"


def service_info(app: ToolShedApp, request_url: URL):
    components = request_url.components
    hostname = components.hostname
    assert hostname
    default_organization_id = ".".join(reversed(hostname.split(".")))
    config = app.config
    organization_id = cast(str, config.ga4gh_service_id or default_organization_id)
    organization_name = cast(str, config.organization_name or organization_id)
    organization_url = cast(str, config.organization_url or f"{components.scheme}://{components.netloc}")

    organization = Organization(
        url=organization_url,
        name=organization_name,
    )
    service_type = ServiceType(
        group="org.ga4gh",
        artifact="trs",
        version="2.1.0",
    )
    extra_kwds = {}
    if environment := config.ga4gh_service_environment:
        extra_kwds["environment"] = environment
    return Service(
        id=organization_id + ".trs",
        name=TRS_SERVICE_NAME,
        description=TRS_SERVICE_DESCRIPTION,
        organization=organization,
        type=service_type,
        version=VERSION,
        **extra_kwds,
    )


def tool_classes() -> list[ToolClass]:
    return [ToolClass(id="galaxy_tool", name="Galaxy Tool", description="Galaxy XML Tools")]


def trs_tool_id_to_guid(trans: ProvidesRepositoriesContext, trs_tool_id: str) -> str:
    guid = decode_identifier(trans.repositories_hostname, trs_tool_id)
    guid = remove_protocol_and_user_from_clone_url(guid)
    return guid


def trs_tool_id_to_repository(trans: ProvidesRepositoriesContext, trs_tool_id: str) -> Repository:
    return guid_to_repository(trans.app, trs_tool_id_to_guid(trans, trs_tool_id))


def get_repository_metadata_by_tool_version(
    app: ToolShedApp, repository: Repository, tool_id: str
) -> dict[str, RepositoryMetadata]:
    versions = {}
    for _, changeset in repository.installable_revisions(app):
        metadata = get_current_repository_metadata_for_changeset_revision(app, repository, changeset)
        tools: Optional[list[dict[str, Any]]] = metadata.metadata.get("tools")
        if not tools:
            continue
        for tool_metadata in tools:
            if tool_metadata["id"] != tool_id:
                continue
            versions[tool_metadata["version"]] = metadata
    return versions


def get_tools_for(repository_metadata: RepositoryMetadata) -> list[dict[str, Any]]:
    tools: Optional[list[dict[str, Any]]] = repository_metadata.metadata.get("tools")
    assert tools
    return tools


def trs_tool_id_to_repository_metadata(
    trans: ProvidesRepositoriesContext, trs_tool_id: str
) -> tuple[Repository, dict[str, RepositoryMetadata]]:
    tool_guid = decode_identifier(trans.repositories_hostname, trs_tool_id)
    tool_guid = remove_protocol_and_user_from_clone_url(tool_guid)
    _, tool_id = tool_guid.rsplit("/", 1)
    repository = guid_to_repository(trans.app, tool_guid)
    app = trans.app
    versions: dict[str, RepositoryMetadata] = get_repository_metadata_by_tool_version(app, repository, tool_id)
    if not versions:
        raise ObjectNotFound()

    return repository, versions


def get_tool(trans: ProvidesRepositoriesContext, trs_tool_id: str) -> Tool:
    guid = decode_identifier(trans.repositories_hostname, trs_tool_id)
    guid = remove_protocol_and_user_from_clone_url(guid)
    repo_metadata = trs_tool_id_to_repository_metadata(trans, trs_tool_id)
    repository, metadata_by_version = repo_metadata

    repo_owner = repository.user.username
    aliases: list[str] = [guid]
    hostname = remove_protocol_and_user_from_clone_url(trans.repositories_hostname)
    url = f"https://{hostname}/repos/{repo_owner}/{repository.name}"

    versions: list[ToolVersion] = []
    for tool_version_str, _ in metadata_by_version.items():
        version_url = url  # TODO:
        tool_version = ToolVersion(
            author=[repo_owner],
            containerfile=False,
            descriptor_type=[DescriptorType.GALAXY],
            id=tool_version_str,
            url=version_url,
            verified=False,
        )
        versions.append(tool_version)
    return Tool(
        aliases=aliases,
        id=trs_tool_id,
        url=url,
        toolclass=tool_classes()[0],
        organization=repo_owner,
        versions=versions,
    )



import logging

from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

column_repository_name = "column_maker_1085"
column_repository_description = "Add column"
column_repository_long_description = "Compute an expression on every row"

convert_repository_name = "convert_chars_1085"
convert_repository_description = "Convert delimiters"
convert_repository_long_description = "Convert delimiters to tab"

category_name = "Test 1085 Advanced Circular Dependencies"
category_description = "Test circular dependency features"

log = logging.getLogger(__name__)


class TestRepositoryDependencies(ShedTwillTestCase):
    """Testing the behavior of repository dependencies with tool panel sections."""

    requires_galaxy = True

    def test_0000_create_or_login_admin_user(self):
        """Create necessary user accounts and login as an admin user."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_create_and_populate_column_repository(self):
        """Create a category for this test suite and add repositories to it."""
        category = self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name=column_repository_name,
            description=column_repository_description,
            long_description=column_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        if self.repository_is_new(repository):
            self.commit_tar_to_repository(
                repository,
                "column_maker/column_maker.tar",
                commit_message="Uploaded column_maker tarball.",
            )

    def test_0010_create_and_populate_convert_repository(self):
        self.login(email=common.admin_email, username=common.admin_username)
        category = self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name=convert_repository_name,
            description=convert_repository_description,
            long_description=convert_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        if self.repository_is_new(repository):
            self.commit_tar_to_repository(
                repository,
                "convert_chars/convert_chars.tar",
                commit_message="Uploaded convert_chars tarball.",
            )

    def test_0015_create_and_upload_dependency_files(self):
        convert_repository = self._get_repository_by_name_and_owner(convert_repository_name, common.test_user_1_name)
        column_repository = self._get_repository_by_name_and_owner(column_repository_name, common.test_user_1_name)
        repository_dependencies_path = self.generate_temp_path("test_1085", additional_paths=["column"])
        repository_tuple = (
            self.url,
            convert_repository.name,
            convert_repository.owner,
            self.get_repository_tip(convert_repository),
        )
        self.create_repository_dependency(
            repository=column_repository, repository_tuples=[repository_tuple], filepath=repository_dependencies_path
        )
        repository_tuple = (
            self.url,
            column_repository.name,
            column_repository.owner,
            self.get_repository_tip(column_repository),
        )
        self.create_repository_dependency(
            repository=convert_repository, repository_tuples=[repository_tuple], filepath=repository_dependencies_path
        )

    def test_0020_install_repositories(self):
        """Install column_maker into column_maker tool panel section and install repository dependencies."""
        self._install_repository(
            column_repository_name,
            common.test_user_1_name,
            category_name,
            install_tool_dependencies=False,
            install_repository_dependencies=True,
            new_tool_panel_section_label="column_maker",
        )
        installed_convert_repository = self._get_installed_repository_by_name_owner(
            convert_repository_name, common.test_user_1_name
        )
        self._refresh_tool_shed_repository(installed_convert_repository)
        self._assert_has_installed_repos_with_names("convert_chars_1085", "column_maker_1085")
        self._assert_is_not_missing_dependency(installed_convert_repository, "column_maker_1085")

    def test_0025_uninstall_column_repository(self):
        """uninstall column_maker, verify same section"""
        installed_column_repository = self._get_installed_repository_by_name_owner(
            column_repository_name, common.test_user_1_name
        )
        self._uninstall_repository(installed_column_repository)
        self.check_galaxy_repository_tool_panel_section(installed_column_repository, "column_maker")

    def test_0030_uninstall_convert_repository(self):
        installed_convert_repository = self._get_installed_repository_by_name_owner(
            convert_repository_name, common.test_user_1_name
        )
        self._uninstall_repository(installed_convert_repository)
        self._refresh_tool_shed_repository(installed_convert_repository)
        self.check_galaxy_repository_tool_panel_section(installed_convert_repository, "column_maker")

    def test_0035_reinstall_column_repository(self):
        """reinstall column_maker into new section 'new_column_maker' (no_changes = false), no dependencies"""
        installed_column_repository = self._get_installed_repository_by_name_owner(
            column_repository_name, common.test_user_1_name
        )
        self.reinstall_repository_api(
            installed_column_repository,
            install_tool_dependencies=False,
            install_repository_dependencies=False,
            new_tool_panel_section_label="new_column_maker",
        )
        self._assert_has_installed_repos_with_names("column_maker_1085")

    def test_0040_reinstall_convert_repository(self):
        """reinstall convert_chars into new section 'new_convert_chars' (no_changes = false), no dependencies"""
        installed_convert_repository = self._get_installed_repository_by_name_owner(
            convert_repository_name, common.test_user_1_name
        )
        self.reinstall_repository_api(
            installed_convert_repository,
            install_tool_dependencies=False,
            install_repository_dependencies=False,
            new_tool_panel_section_label="new_convert_chars",
        )
        self._assert_has_installed_repos_with_names("convert_chars_1085")

    # The following check fails somewhere around 5% of the time maybe on Jenkins.
    # https://jenkins.galaxyproject.org/job/docker-toolshed/5578/
    # https://jenkins.galaxyproject.org/job/docker-toolshed/5198/
    # def test_0045_uninstall_and_verify_tool_panel_sections( self ):
    #    '''uninstall both and verify tool panel sections'''
    #    installed_convert_repository = self._get_installed_repository_by_name_owner( convert_repository_name,
    #                                                                                        common.test_user_1_name )
    #    installed_column_repository = self._get_installed_repository_by_name_owner( column_repository_name,
    #                                                                                        common.test_user_1_name )
    #    self._uninstall_repository( installed_convert_repository )
    #    self._uninstall_repository( installed_column_repository )
    #    self.test_db_util.ga_refresh( installed_convert_repository )
    #    self.test_db_util.ga_refresh( installed_column_repository )
    #    self.check_galaxy_repository_tool_panel_section( installed_column_repository, 'new_column_maker' )
    #    self.check_galaxy_repository_tool_panel_section( installed_convert_repository, 'new_convert_chars' )



"""
Middleware for handling $REMOTE_USER if use_remote_user is enabled.
"""

import socket

from galaxy.util import safe_str_cmp

errorpage = """
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" "http://www.w3.org/TR/html4/loose.dtd">
<html lang="en">
    <head>
        <title>Galaxy</title>
        <style type="text/css">
        body {
            min-width: 500px;
            text-align: center;
        }
        .errormessage {
            font: 75%% verdana, "Bitstream Vera Sans", geneva, arial, helvetica, helve, sans-serif;
            padding: 10px;
            margin: 100px auto;
            min-height: 32px;
            max-width: 500px;
            border: 1px solid #AA6666;
            background-color: #FFCCCC;
            text-align: left;
        }
        </style>
    </head>
    <body>
        <div class="errormessage">
            <h4>%s</h4>
            <p>%s</p>
        </div>
    </body>
</html>
"""


class RemoteUser:
    def __init__(self, app, maildomain=None, display_servers=None, admin_users=None, remote_user_secret_header=None):
        self.app = app
        self.maildomain = maildomain
        self.display_servers = display_servers or []
        self.admin_users = admin_users or []
        self.config_secret_header = remote_user_secret_header

    def __call__(self, environ, start_response):
        environ["webapp"] = "tool_shed"
        # Allow display servers
        if self.display_servers and "REMOTE_ADDR" in environ:
            try:
                host = socket.gethostbyaddr(environ["REMOTE_ADDR"])[0]
            except (OSError, socket.herror, socket.gaierror, socket.timeout):
                # in the event of a lookup failure, deny access
                host = None
            if host in self.display_servers:
                environ["HTTP_REMOTE_USER"] = "remote_display_server@%s" % (self.maildomain or "example.org")
                return self.app(environ, start_response)

        # If the secret header is enabled, we expect upstream to send along some key
        # in HTTP_GX_SECRET, so we'll need to compare that here to the correct value
        #
        # This is not an ideal location for this function.  The reason being
        # that because this check is done BEFORE the REMOTE_USER check,  it is
        # possible to attack the GX_SECRET key without having correct
        # credentials. However, that's why it's not "ideal", but it is "good
        # enough". The only users able to exploit this are ones with access to
        # the local system (unless Galaxy is listening on 0.0.0.0....). It
        # seems improbable that an attacker with access to the server hosting
        # Galaxy would not have access to Galaxy itself, and be attempting to
        # attack the system
        if self.config_secret_header is not None:
            if not safe_str_cmp(environ.get("HTTP_GX_SECRET"), self.config_secret_header):
                title = "Access to Galaxy is denied"
                message = """
                Galaxy is configured to authenticate users via an external
                method (such as HTTP authentication in Apache), but an
                incorrect shared secret key was provided by the
                upstream (proxy) server.</p>
                <p>Please contact your local Galaxy administrator.  The
                variable <code>remote_user_secret</code> and
                <code>GX_SECRET</code> header must be set before you may
                access Galaxy.
                """
                return self.error(start_response, title, message)

        # Apache sets REMOTE_USER to the string '(null)' when using the Rewrite* method for passing REMOTE_USER and a user is
        # un-authenticated.  Any other possible values need to go here as well.
        path_info = environ.get("PATH_INFO", "")
        if "HTTP_REMOTE_USER" in environ and environ["HTTP_REMOTE_USER"] != "(null)":
            if not environ["HTTP_REMOTE_USER"].count("@"):
                if self.maildomain is not None:
                    environ["HTTP_REMOTE_USER"] += f"@{self.maildomain}"
                else:
                    title = "Access to this Galaxy tool shed is denied"
                    message = """
                        This Galaxy tool shed is configured to authenticate users via an external
                        method (such as HTTP authentication in Apache), but only a username (not
                        an email address) was provided by the upstream (proxy) server.  Since tool
                        shed usernames are email addresses, a default mail domain must be set.</[>
                        <p>The variable <code>remote_user_maildomain</code> must be set before you
                        can access this tool shed.  Contact your local tool shed administrator.
                    """
                    return self.error(start_response, title, message)
            return self.app(environ, start_response)
        elif path_info.startswith("/api/"):
            # The API handles its own authentication via keys
            return self.app(environ, start_response)
        elif path_info.startswith("/user/api_keys"):
            # api_keys can be managed when remote_user is in use.
            pass
        else:
            title = "Access to this Galaxy tool shed is denied"
            message = """
                This Galaxy tool shed is configured to authenticate users via an external
                method (such as HTTP authentication in Apache), but a username was not
                provided by the upstream (proxy) server.  This is generally due to a
                misconfiguration in the upstream server.</p>
                <p>Contact your local Galaxy tool shed administrator.
            """
            return self.error(start_response, title, message)

    def error(
        self, start_response, title="Access denied", message="Contact your local Galaxy tool shed administrator."
    ):
        start_response("403 Forbidden", [("Content-type", "text/html")])
        return [errorpage % (title, message)]



"""
Universe configuration builder.
"""

import configparser
import logging
import logging.config
import os
from datetime import timedelta

from galaxy.config import (
    BaseAppConfiguration,
    CommonConfigurationMixin,
    expand_pretty_datetime_format,
    get_database_engine_options,
    TOOL_SHED_CONFIG_SCHEMA_PATH,
)
from galaxy.config.schema import AppSchema
from galaxy.exceptions import ConfigurationError
from galaxy.util import string_as_bool
from galaxy.version import (
    VERSION,
    VERSION_MAJOR,
    VERSION_MINOR,
)

log = logging.getLogger(__name__)

TOOLSHED_APP_NAME = "tool_shed"
SHED_API_VERSION = os.environ.get("TOOL_SHED_API_VERSION", "v1")


class ToolShedAppConfiguration(BaseAppConfiguration, CommonConfigurationMixin):
    default_config_file_name = "tool_shed.yml"

    add_sample_file_to_defaults = {"datatypes_config_file"}

    def _load_schema(self):
        return AppSchema(TOOL_SHED_CONFIG_SCHEMA_PATH, TOOLSHED_APP_NAME)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._process_config(kwargs)

    @property
    def shed_tool_data_path(self):
        return self.tool_data_path

    def check(self):
        # Check that required directories exist; attempt to create otherwise
        paths_to_check = [
            self.file_path,
            self.hgweb_config_dir,
            self.tool_data_path,
            self.template_cache_path,
            os.path.join(self.tool_data_path, "shared", "jars"),
        ]
        for path in paths_to_check:
            self._ensure_directory(path)
        # Check that required files exist.
        if not os.path.isfile(self.datatypes_config):
            raise ConfigurationError(f"File not found: {self.datatypes_config}")

    def _process_config(self, kwargs):
        # Backwards compatibility for names used in too many places to fix
        self.datatypes_config = self.datatypes_config_file
        # Collect the umask and primary gid from the environment
        self.umask = os.umask(0o77)  # get the current umask
        os.umask(self.umask)  # can't get w/o set, so set it back
        self.gid = os.getgid()  # if running under newgrp(1) we'll need to fix the group of data created on the cluster
        self.version_major = VERSION_MAJOR
        self.version_minor = VERSION_MINOR
        self.version = VERSION
        # Database related configuration
        if not self.database_connection:  # Provide default if not supplied by user
            self.database_connection = f"sqlite:///{self._in_data_dir('community.sqlite')}?isolation_level=IMMEDIATE"
        self.database_engine_options = get_database_engine_options(kwargs)
        self.database_create_tables = string_as_bool(kwargs.get("database_create_tables", "True"))
        # Where dataset files are stored
        self.file_path = self._in_root_dir(self.file_path)
        self.new_file_path = self._in_root_dir(self.new_file_path)
        self.cookie_path = kwargs.get("cookie_path")
        self.cookie_domain = kwargs.get("cookie_domain")
        self.enable_quotas = string_as_bool(kwargs.get("enable_quotas", False))
        # Tool stuff
        self.tool_path = self._in_root_dir(kwargs.get("tool_path", "tools"))
        self.tool_secret = kwargs.get("tool_secret", "")
        self.tool_data_path = os.path.join(os.getcwd(), kwargs.get("tool_data_path", "shed-tool-data"))
        self.tool_data_table_config_path = None
        self.integrated_tool_panel_config = self._in_root_dir(
            kwargs.get("integrated_tool_panel_config", "integrated_tool_panel.xml")
        )
        self.builds_file_path = self._in_root_dir(
            kwargs.get("builds_file_path", os.path.join(self.tool_data_path, "shared", "ucsc", "builds.txt"))
        )
        self.len_file_path = self._in_root_dir(
            kwargs.get("len_file_path", os.path.join(self.tool_data_path, "shared", "ucsc", "chrom"))
        )
        self.ftp_upload_dir = kwargs.get("ftp_upload_dir")
        self.update_integrated_tool_panel = False
        # Galaxy flavor Docker Image
        self.user_activation_on = None
        self.registration_warning_message = kwargs.get("registration_warning_message")
        self.email_domain_blocklist_content = None
        self.email_domain_allowlist_content = None
        self.template_cache_path = self._in_root_dir(
            kwargs.get("template_cache_path", "database/compiled_templates/community")
        )
        self.error_email_to = kwargs.get("error_email_to")
        self.pretty_datetime_format = expand_pretty_datetime_format(self.pretty_datetime_format)
        # Configuration for the message box directly below the masthead.
        self.wiki_url = kwargs.get("wiki_url", "https://galaxyproject.org/")
        self.blog_url = kwargs.get("blog_url")
        self.screencasts_url = kwargs.get("screencasts_url")
        self.log_events = False
        self.cloud_controller_instance = False
        self.server_name = ""
        # Where the tool shed hgweb.config file is stored - the default is the Galaxy installation directory.
        self.hgweb_config_dir = self._in_root_dir(self.hgweb_config_dir) or self.root
        # Proxy features
        self.drmaa_external_runjob_script = kwargs.get("drmaa_external_runjob_script")
        # Parse global_conf and save the parser
        global_conf = kwargs.get("global_conf")
        global_conf_parser = configparser.ConfigParser()
        self.global_conf_parser = global_conf_parser
        if global_conf and "__file__" in global_conf and ".yml" not in global_conf["__file__"]:
            global_conf_parser.read(global_conf["__file__"])
        self.running_functional_tests = string_as_bool(kwargs.get("running_functional_tests", False))
        self.citation_cache_data_dir = self._in_root_dir(
            kwargs.get("citation_cache_data_dir", "database/tool_shed_citations/data")
        )
        self.citation_cache_lock_dir = self._in_root_dir(
            kwargs.get("citation_cache_lock_dir", "database/tool_shed_citations/locks")
        )
        self.citation_cache_url = kwargs.get("citation_cache_lock_dir", None)
        self.citation_cache_schema_name = kwargs.get("citation_cache_schema_name", None)
        self.citation_cache_table_name = kwargs.get("citation_cache_table_name", None)
        self.password_expiration_period = timedelta(days=int(self.password_expiration_period))

        # Security/Policy Compliance
        self.redact_username_during_deletion = False
        self.redact_email_during_deletion = False
        self.redact_username_in_logs = False
        self.enable_beta_gdpr = string_as_bool(kwargs.get("enable_beta_gdpr", False))
        if self.enable_beta_gdpr:
            self.redact_username_during_deletion = True
            self.redact_email_during_deletion = True
            self.redact_username_in_logs = True
            self.allow_user_deletion = True


Configuration = ToolShedAppConfiguration



from time import (
    gmtime,
    strftime,
)

from sqlalchemy import select

from tool_shed.webapp.model import Repository


class ShedCounter:
    def __init__(self, model):
        # TODO: Enhance the ShedCounter to retrieve information from the db instead of displaying what's currently in memory.
        self.model = model
        self.custom_datatypes = 0
        self.generation_time = strftime("%b %d, %Y", gmtime())
        self.deleted_repositories = 0
        self.deprecated_repositories = 0
        self.invalid_versions_of_tools = 0
        self.repositories = 0
        self.total_clones = 0
        self.valid_versions_of_tools = 0
        self.unique_owners = 0
        self.unique_valid_tools = 0
        self.workflows = 0

    def generate_statistics(self, sa_session):
        self.custom_datatypes = 0
        self.deleted_repositories = 0
        self.deprecated_repositories = 0
        self.invalid_versions_of_tools = 0
        self.repositories = 0
        self.total_clones = 0
        self.unique_owners = 0
        self.valid_versions_of_tools = 0
        self.unique_valid_tools = 0
        self.workflows = 0
        unique_user_ids = []
        for repository in sa_session.scalars(select(Repository)):
            self.repositories += 1
            self.total_clones += repository.times_downloaded
            is_deleted = repository.deleted
            if is_deleted:
                self.deleted_repositories += 1
            else:
                if repository.deprecated:
                    self.deprecated_repositories += 1
                if repository.user_id not in unique_user_ids:
                    self.unique_owners += 1
                    unique_user_ids.append(repository.user_id)
                processed_datatypes = []
                processed_guids = []
                processed_invalid_tool_configs = []
                processed_relative_workflow_paths = []
                processed_tool_ids = []
                # A repository's metadata_revisions are those that ignore the value of the
                # repository_metadata.downloadable column.
                for metadata_revision in repository.metadata_revisions:
                    metadata = metadata_revision.metadata
                    if "tools" in metadata:
                        tool_dicts = metadata["tools"]
                        for tool_dict in tool_dicts:
                            if "guid" in tool_dict:
                                guid = tool_dict["guid"]
                                if guid not in processed_guids:
                                    self.valid_versions_of_tools += 1
                                    processed_guids.append(guid)
                            if "id" in tool_dict:
                                tool_id = tool_dict["id"]
                                if tool_id not in processed_tool_ids:
                                    self.unique_valid_tools += 1
                                    processed_tool_ids.append(tool_id)
                    if "invalid_tools" in metadata:
                        invalid_tool_configs = metadata["invalid_tools"]
                        for invalid_tool_config in invalid_tool_configs:
                            if invalid_tool_config not in processed_invalid_tool_configs:
                                self.invalid_versions_of_tools += 1
                                processed_invalid_tool_configs.append(invalid_tool_config)
                    if "datatypes" in metadata:
                        datatypes = metadata["datatypes"]
                        for datatypes_dict in datatypes:
                            if "extension" in datatypes_dict:
                                extension = datatypes_dict["extension"]
                                if extension not in processed_datatypes:
                                    self.custom_datatypes += 1
                                    processed_datatypes.append(extension)
                    if "workflows" in metadata:
                        workflows = metadata["workflows"]
                        for workflow_tup in workflows:
                            relative_path, exported_workflow_dict = workflow_tup
                            if relative_path not in processed_relative_workflow_paths:
                                self.workflows += 1
                                processed_relative_workflow_paths.append(relative_path)
        self.generation_time = strftime("%b %d, %Y", gmtime())



import logging

from fastapi import (
    Path,
    Request,
    Response,
)

from galaxy.tool_util.parameters import (
    RequestToolState,
    to_json_schema_string,
)
from tool_shed.context import SessionRequestContext
from tool_shed.managers.tools import (
    parsed_tool_model_cached_for,
    search,
)
from tool_shed.managers.trs import (
    get_tool,
    service_info,
    tool_classes,
)
from tool_shed.structured_app import ToolShedApp
from tool_shed.util.shed_index import build_index
from tool_shed_client.schema import (
    BuildSearchIndexResponse,
    ShedParsedTool,
)
from tool_shed_client.schema.trs import (
    Tool,
    ToolClass,
    ToolVersion,
)
from tool_shed_client.schema.trs_service_info import Service
from . import (
    depends,
    DependsOnTrans,
    RepositorySearchPageSizeQueryParam,
    Router,
    ToolSearchPageQueryParam,
    ToolsIndexQueryParam,
)

log = logging.getLogger(__name__)

router = Router(tags=["tools"])

TOOL_ID_PATH_PARAM: str = Path(
    ...,
    title="GA4GH TRS Tool ID",
    description="See also https://ga4gh.github.io/tool-registry-service-schemas/DataModel/#trs-tool-and-trs-tool-version-ids",
)

TOOL_VERSION_PATH_PARAM: str = Path(
    ...,
    title="Galaxy Tool Wrapper Version",
    description="The full version string defined on the Galaxy tool wrapper.",
)


def json_schema_response(pydantic_model) -> Response:
    json_str = to_json_schema_string(pydantic_model)
    return Response(content=json_str, media_type="application/json")


@router.cbv
class FastAPITools:
    app: ToolShedApp = depends(ToolShedApp)

    @router.get(
        "/api/tools",
        operation_id="tools__index",
    )
    def index(
        self,
        q: str = ToolsIndexQueryParam,
        page: int = ToolSearchPageQueryParam,
        page_size: int = RepositorySearchPageSizeQueryParam,
        trans: SessionRequestContext = DependsOnTrans,
    ):
        search_results = search(trans, q, page, page_size)
        return search_results

    @router.put(
        "/api/tools/build_search_index",
        operation_id="tools__build_search_index",
        require_admin=True,
    )
    def build_search_index(self) -> BuildSearchIndexResponse:
        """Not part of the stable API, just something to simplify
        bootstrapping tool sheds, scripting, testing, etc...
        """
        config = self.app.config
        repos_indexed, tools_indexed = build_index(
            config.whoosh_index_dir,
            config.file_path,
            config.hgweb_config_dir,
            config.hgweb_repo_prefix,
            config.database_connection,
        )
        return BuildSearchIndexResponse(
            repositories_indexed=repos_indexed,
            tools_indexed=tools_indexed,
        )

    @router.get("/api/ga4gh/trs/v2/service-info", operation_id="tools_trs_service_info")
    def service_info(self, request: Request) -> Service:
        return service_info(self.app, request.url)

    @router.get("/api/ga4gh/trs/v2/toolClasses", operation_id="tools__trs_tool_classes")
    def tool_classes(self) -> list[ToolClass]:
        return tool_classes()

    @router.get(
        "/api/ga4gh/trs/v2/tools",
        operation_id="tools__trs_index",
    )
    def trs_index(
        self,
    ):
        # we probably want to be able to query the database at the
        # tool level and such to do this right?
        return []

    @router.get(
        "/api/ga4gh/trs/v2/tools/{tool_id}",
        operation_id="tools__trs_get",
    )
    def trs_get(
        self,
        trans: SessionRequestContext = DependsOnTrans,
        tool_id: str = TOOL_ID_PATH_PARAM,
    ) -> Tool:
        return get_tool(trans, tool_id)

    @router.get(
        "/api/ga4gh/trs/v2/tools/{tool_id}/versions",
        operation_id="tools__trs_get_versions",
    )
    def trs_get_versions(
        self,
        trans: SessionRequestContext = DependsOnTrans,
        tool_id: str = TOOL_ID_PATH_PARAM,
    ) -> list[ToolVersion]:
        return get_tool(trans, tool_id).versions

    @router.get(
        "/api/tools/{tool_id}/versions/{tool_version}",
        operation_id="tools__parameter_model",
        summary="Return Galaxy's meta model description of the tool's inputs",
    )
    def show_tool(
        self,
        trans: SessionRequestContext = DependsOnTrans,
        tool_id: str = TOOL_ID_PATH_PARAM,
        tool_version: str = TOOL_VERSION_PATH_PARAM,
    ) -> ShedParsedTool:
        return parsed_tool_model_cached_for(trans, tool_id, tool_version)

    @router.get(
        "/api/tools/{tool_id}/versions/{tool_version}/parameter_request_schema",
        operation_id="tools__parameter_request_model",
        summary="Return a JSON schema description of the tool's inputs for the tool request API that will be added to Galaxy at some point",
        description="The tool request schema includes validation of map/reduce concepts that can be consumed by the tool execution API and not just the request for a single execution.",
    )
    def tool_state(
        self,
        trans: SessionRequestContext = DependsOnTrans,
        tool_id: str = TOOL_ID_PATH_PARAM,
        tool_version: str = TOOL_VERSION_PATH_PARAM,
    ) -> Response:
        parsed_tool = parsed_tool_model_cached_for(trans, tool_id, tool_version)
        return json_schema_response(RequestToolState.parameter_model_for(parsed_tool.inputs))



import logging
import threading

from galaxy import util
from galaxy.tool_shed.util import utility_container_manager
from galaxy.tool_shed.util.container_util import (
    generate_repository_dependencies_key_for_repository,
    STRSEP,
)
from galaxy.util import UNKNOWN
from galaxy.util.tool_shed.common_util import parse_repository_dependency_tuple
from tool_shed.util.readme_util import build_readme_files_dict

log = logging.getLogger(__name__)


class FailedTest:
    """Failed tool tests object"""

    def __init__(self, id=None, stderr=None, test_id=None, tool_id=None, tool_version=None, traceback=None):
        self.id = id
        self.stderr = stderr
        self.test_id = test_id
        self.tool_id = tool_id
        self.tool_version = tool_version
        self.traceback = traceback


class InvalidRepositoryDependency:
    """Invalid repository dependency definition object"""

    def __init__(
        self,
        id=None,
        toolshed=None,
        repository_name=None,
        repository_owner=None,
        changeset_revision=None,
        prior_installation_required=False,
        only_if_compiling_contained_td=False,
        error=None,
    ):
        self.id = id
        self.toolshed = toolshed
        self.repository_name = repository_name
        self.repository_owner = repository_owner
        self.changeset_revision = changeset_revision
        self.prior_installation_required = prior_installation_required
        self.only_if_compiling_contained_td = only_if_compiling_contained_td
        self.error = error


class InvalidToolDependency:
    """Invalid tool dependency definition object"""

    def __init__(self, id=None, name=None, version=None, type=None, error=None):
        self.id = id
        self.name = name
        self.version = version
        self.type = type
        self.error = error


class MissingTestComponent:
    """Missing tool test components object"""

    def __init__(self, id=None, missing_components=None, tool_guid=None, tool_id=None, tool_version=None):
        self.id = id
        self.missing_components = missing_components
        self.tool_guid = tool_guid
        self.tool_id = tool_id
        self.tool_version = tool_version


class NotTested:
    """NotTested object"""

    def __init__(self, id=None, reason=None):
        self.id = id
        self.reason = reason


class PassedTest:
    """Passed tool tests object"""

    def __init__(self, id=None, test_id=None, tool_id=None, tool_version=None):
        self.id = id
        self.test_id = test_id
        self.tool_id = tool_id
        self.tool_version = tool_version


class RepositoryInstallationError:
    """Repository installation error object"""

    def __init__(self, id=None, tool_shed=None, name=None, owner=None, changeset_revision=None, error_message=None):
        self.id = id
        self.tool_shed = tool_shed
        self.name = name
        self.owner = owner
        self.changeset_revision = changeset_revision
        self.error_message = error_message


class RepositorySuccessfulInstallation:
    """Repository installation object"""

    def __init__(self, id=None, tool_shed=None, name=None, owner=None, changeset_revision=None):
        self.id = id
        self.tool_shed = tool_shed
        self.name = name
        self.owner = owner
        self.changeset_revision = changeset_revision


class ToolDependencyInstallationError:
    """Tool dependency installation error object"""

    def __init__(self, id=None, type=None, name=None, version=None, error_message=None):
        self.id = id
        self.type = type
        self.name = name
        self.version = version
        self.error_message = error_message


class ToolDependencySuccessfulInstallation:
    """Tool dependency installation object"""

    def __init__(self, id=None, type=None, name=None, version=None, installation_directory=None):
        self.id = id
        self.type = type
        self.name = name
        self.version = version
        self.installation_directory = installation_directory


class ToolShedUtilityContainerManager(utility_container_manager.UtilityContainerManager):
    def __init__(self, app):
        self.app = app

    def build_invalid_repository_dependencies_root_folder(self, folder_id, invalid_repository_dependencies_dict):
        """Return a folder hierarchy containing invalid repository dependencies."""
        label = "Invalid repository dependencies"
        if invalid_repository_dependencies_dict:
            invalid_repository_dependency_id = 0
            folder_id += 1
            invalid_repository_dependencies_root_folder = utility_container_manager.Folder(
                id=folder_id, key="root", label="root", parent=None
            )
            folder_id += 1
            invalid_repository_dependencies_folder = utility_container_manager.Folder(
                id=folder_id,
                key="invalid_repository_dependencies",
                label=label,
                parent=invalid_repository_dependencies_root_folder,
            )
            invalid_repository_dependencies_root_folder.folders.append(invalid_repository_dependencies_folder)
            invalid_repository_dependencies = invalid_repository_dependencies_dict["repository_dependencies"]
            for invalid_repository_dependency in invalid_repository_dependencies:
                folder_id += 1
                invalid_repository_dependency_id += 1
                (
                    toolshed,
                    name,
                    owner,
                    changeset_revision,
                    prior_installation_required,
                    only_if_compiling_contained_td,
                    error,
                ) = parse_repository_dependency_tuple(invalid_repository_dependency, contains_error=True)
                key = generate_repository_dependencies_key_for_repository(
                    toolshed,
                    name,
                    owner,
                    changeset_revision,
                    prior_installation_required,
                    only_if_compiling_contained_td,
                )
                label = f"Repository <b>{name}</b> revision <b>{changeset_revision}</b> owned by <b>{owner}</b>"
                folder = utility_container_manager.Folder(
                    id=folder_id, key=key, label=label, parent=invalid_repository_dependencies_folder
                )
                ird = InvalidRepositoryDependency(
                    id=invalid_repository_dependency_id,
                    toolshed=toolshed,
                    repository_name=name,
                    repository_owner=owner,
                    changeset_revision=changeset_revision,
                    prior_installation_required=util.asbool(prior_installation_required),
                    only_if_compiling_contained_td=util.asbool(only_if_compiling_contained_td),
                    error=error,
                )
                folder.invalid_repository_dependencies.append(ird)
                invalid_repository_dependencies_folder.folders.append(folder)
        else:
            invalid_repository_dependencies_root_folder = None
        return folder_id, invalid_repository_dependencies_root_folder

    def build_invalid_tool_dependencies_root_folder(self, folder_id, invalid_tool_dependencies_dict):
        """Return a folder hierarchy containing invalid tool dependencies."""
        # # INvalid tool dependencies are always packages like:
        # {"R/2.15.1": {"name": "R", "readme": "some string", "type": "package", "version": "2.15.1" "error" : "some sting" }
        label = "Invalid tool dependencies"
        if invalid_tool_dependencies_dict:
            invalid_tool_dependency_id = 0
            folder_id += 1
            invalid_tool_dependencies_root_folder = utility_container_manager.Folder(
                id=folder_id, key="root", label="root", parent=None
            )
            folder_id += 1
            invalid_tool_dependencies_folder = utility_container_manager.Folder(
                id=folder_id, key="invalid_tool_dependencies", label=label, parent=invalid_tool_dependencies_root_folder
            )
            invalid_tool_dependencies_root_folder.folders.append(invalid_tool_dependencies_folder)
            for requirements_dict in invalid_tool_dependencies_dict.values():
                folder_id += 1
                invalid_tool_dependency_id += 1
                try:
                    name = requirements_dict["name"]
                    type = requirements_dict["type"]
                    version = requirements_dict["version"]
                    error = requirements_dict["error"]
                except Exception as e:
                    name = UNKNOWN
                    type = UNKNOWN
                    version = UNKNOWN
                    error = str(e)
                key = self.generate_tool_dependencies_key(name, version, type)
                label = f"Version <b>{version}</b> of the <b>{name}</b> <b>{type}</b>"
                folder = utility_container_manager.Folder(
                    id=folder_id, key=key, label=label, parent=invalid_tool_dependencies_folder
                )
                itd = InvalidToolDependency(
                    id=invalid_tool_dependency_id, name=name, version=version, type=type, error=error
                )
                folder.invalid_tool_dependencies.append(itd)
                invalid_tool_dependencies_folder.folders.append(folder)
        else:
            invalid_tool_dependencies_root_folder = None
        return folder_id, invalid_tool_dependencies_root_folder

    def build_repository_containers(
        self, repository, changeset_revision, repository_dependencies, repository_metadata, exclude=None
    ):
        """
        Return a dictionary of containers for the received repository's dependencies and
        contents for display in the Tool Shed.
        """
        if exclude is None:
            exclude = []
        containers_dict = dict(
            datatypes=None,
            invalid_tools=None,
            readme_files=None,
            repository_dependencies=None,
            tool_dependencies=None,
            valid_tools=None,
            valid_data_managers=None,
        )
        if repository_metadata:
            metadata = repository_metadata.metadata
            lock = threading.Lock()
            lock.acquire(True)
            try:
                folder_id = 0
                # Invalid repository dependencies container.
                if metadata:
                    if (
                        "invalid_repository_dependencies" not in exclude
                        and "invalid_repository_dependencies" in metadata
                    ):
                        invalid_repository_dependencies = metadata["invalid_repository_dependencies"]
                        (
                            folder_id,
                            invalid_repository_dependencies_root_folder,
                        ) = self.build_invalid_repository_dependencies_root_folder(
                            folder_id, invalid_repository_dependencies
                        )
                        containers_dict["invalid_repository_dependencies"] = invalid_repository_dependencies_root_folder
                # Invalid tool dependencies container.
                if metadata:
                    if "invalid_tool_dependencies" not in exclude and "invalid_tool_dependencies" in metadata:
                        invalid_tool_dependencies = metadata["invalid_tool_dependencies"]
                        (
                            folder_id,
                            invalid_tool_dependencies_root_folder,
                        ) = self.build_invalid_tool_dependencies_root_folder(folder_id, invalid_tool_dependencies)
                        containers_dict["invalid_tool_dependencies"] = invalid_tool_dependencies_root_folder
                # Invalid tools container.
                if metadata:
                    if "invalid_tools" not in exclude and "invalid_tools" in metadata:
                        invalid_tool_configs = metadata["invalid_tools"]
                        folder_id, invalid_tools_root_folder = self.build_invalid_tools_folder(
                            folder_id,
                            invalid_tool_configs,
                            changeset_revision,
                            repository=repository,
                            label="Invalid tools",
                        )
                        containers_dict["invalid_tools"] = invalid_tools_root_folder
                # Readme files container.
                if metadata:
                    if "readme_files" not in exclude and "readme_files" in metadata:
                        readme_files_dict = build_readme_files_dict(self.app, repository, changeset_revision, metadata)
                        folder_id, readme_files_root_folder = self.build_readme_files_folder(
                            folder_id, readme_files_dict
                        )
                        containers_dict["readme_files"] = readme_files_root_folder
                if "repository_dependencies" not in exclude:
                    # Repository dependencies container.
                    folder_id, repository_dependencies_root_folder = self.build_repository_dependencies_folder(
                        folder_id=folder_id,
                        repository_dependencies=repository_dependencies,
                        label="Repository dependencies",
                        installed=False,
                    )
                    if repository_dependencies_root_folder:
                        containers_dict["repository_dependencies"] = repository_dependencies_root_folder
                # Tool dependencies container.
                if metadata:
                    if "tool_dependencies" not in exclude and "tool_dependencies" in metadata:
                        tool_dependencies = metadata["tool_dependencies"]
                        if "orphan_tool_dependencies" in metadata:
                            # The use of the orphan_tool_dependencies category in metadata has been deprecated,
                            # but we still need to check in case the metadata is out of date.
                            orphan_tool_dependencies = metadata["orphan_tool_dependencies"]
                            tool_dependencies.update(orphan_tool_dependencies)
                        # Tool dependencies can be categorized as orphans only if the repository contains tools.
                        if "tools" not in exclude:
                            tools = metadata.get("tools", [])
                            tools.extend(metadata.get("invalid_tools", []))
                        folder_id, tool_dependencies_root_folder = self.build_tool_dependencies_folder(
                            folder_id, tool_dependencies, missing=False, new_install=False
                        )
                        containers_dict["tool_dependencies"] = tool_dependencies_root_folder
                # Valid tools container.
                if metadata:
                    if "tools" not in exclude and "tools" in metadata:
                        valid_tools = metadata["tools"]
                        folder_id, valid_tools_root_folder = self.build_tools_folder(
                            folder_id, valid_tools, repository, changeset_revision, label="Valid tools"
                        )
                        containers_dict["valid_tools"] = valid_tools_root_folder
                # Valid Data Managers container
                if metadata:
                    if "data_manager" not in exclude and "data_manager" in metadata:
                        data_managers = metadata["data_manager"].get("data_managers", None)
                        folder_id, data_managers_root_folder = self.build_data_managers_folder(
                            folder_id, data_managers, label="Data Managers"
                        )
                        containers_dict["valid_data_managers"] = data_managers_root_folder
                        error_messages = metadata["data_manager"].get("error_messages", None)
                        data_managers = metadata["data_manager"].get("invalid_data_managers", None)
                        folder_id, data_managers_root_folder = self.build_invalid_data_managers_folder(
                            folder_id, data_managers, error_messages, label="Invalid Data Managers"
                        )
                        containers_dict["invalid_data_managers"] = data_managers_root_folder
            except Exception:
                log.exception("Exception in build_repository_containers")
            finally:
                lock.release()
        return containers_dict

    def generate_tool_dependencies_key(self, name, version, type):
        return f"{str(name)}{STRSEP}{str(version)}{STRSEP}{str(type)}"
